/*
 * This file is part of TechReborn, licensed under the MIT License (MIT).
 *
 * Copyright (c) 2018 TechReborn
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package techreborn.blockentity.fusionReactor;

import reborncore.api.IToolDrop;
import reborncore.api.blockentity.InventoryProvider;
import reborncore.client.containerBuilder.IContainerProvider;
import reborncore.client.containerBuilder.builder.BuiltContainer;
import reborncore.client.containerBuilder.builder.ContainerBuilder;
import reborncore.common.crafting.RebornRecipe;
import reborncore.common.crafting.ingredient.RebornIngredient;
import reborncore.common.powerSystem.PowerAcceptorBlockEntity;
import reborncore.common.util.ItemUtils;
import reborncore.common.util.RebornInventory;
import reborncore.common.util.Torus;
import techreborn.api.recipe.recipes.FusionReactorRecipe;
import techreborn.config.TechRebornConfig;
import techreborn.init.ModRecipes;
import techreborn.init.TRBlockEntities;
import techreborn.init.TRContent;

import java.util.List;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;

public class FusionControlComputerBlockEntity extends PowerAcceptorBlockEntity
		implements IToolDrop, InventoryProvider, IContainerProvider {

	public RebornInventory<FusionControlComputerBlockEntity> inventory;

	public int coilCount = 0;
	public int crafingTickTime = 0;
	public int finalTickTime = 0;
	public int neededPower = 0;
	public int size = 6;
	public int state = -1;
	int topStackSlot = 0;
	int bottomStackSlot = 1;
	int outputStackSlot = 2;
	FusionReactorRecipe currentRecipe = null;
	boolean hasStartedCrafting = false;
	long lastTick = -1;

	public FusionControlComputerBlockEntity() {
		super(TRBlockEntities.FUSION_CONTROL_COMPUTER);
		checkOverfill = false;
		this.inventory = new RebornInventory<>(3, "FusionControlComputerBlockEntity", 64, this);
	}

	/**
	 * Check that reactor has all necessary coils in place
	 * 
	 * @return boolean Return true if coils are present
	 */
	public boolean checkCoils() {
		List<class_2338> coils = Torus.generate(field_11867, size);
		for(class_2338 coilPos : coils){
			if (!isCoil(coilPos)) {
				coilCount = 0;
				return false;
			}
		}
		coilCount = coils.size();
		return true;
	}

	/**
	 * Checks if block is fusion coil
	 * 
	 * @param pos coordinate for block
	 * @return boolean Returns true if block is fusion coil
	 */
	public boolean isCoil(class_2338 pos) {
		return field_11863.method_8320(pos).method_11614() == TRContent.Machine.FUSION_COIL.block;
	}

	/**
	 * Resets crafter progress and recipe
	 */
	private void resetCrafter() {
		currentRecipe = null;
		crafingTickTime = 0;
		finalTickTime = 0;
		neededPower = 0;
		hasStartedCrafting = false;
	}

	/**
	 * Checks that ItemStack could be inserted into slot provided, including check
	 * for existing item in slot and maximum stack size
	 * 
	 * @param stack ItemStack ItemStack to insert
	 * @param slot int Slot ID to check
	 * @param tags boolean Should we use tags
	 * @return boolean Returns true if ItemStack will fit into slot
	 */
	public boolean canFitStack(class_1799 stack, int slot, boolean tags) {// Checks to see if it can
																								// fit the stack
		if (stack.method_7960()) {
			return true;
		}
		if (inventory.method_5438(slot).method_7960()) {
			return true;
		}
		if (ItemUtils.isItemEqual(inventory.method_5438(slot), stack, true, tags)) {
			if (stack.method_7947() + inventory.method_5438(slot).method_7947() <= stack.method_7914()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns progress scaled to input value
	 * 
	 * @param scale int Maximum value for progress
	 * @return int Scale of progress
	 */
	public int getProgressScaled(int scale) {
		if (crafingTickTime != 0 && finalTickTime != 0) {
			return crafingTickTime * scale / finalTickTime;
		}
		return 0;
	}

	/**
	 * Tries to set current recipe based in inputs in reactor
	 */
	private void updateCurrentRecipe() {
		for (RebornRecipe recipe : ModRecipes.FUSION_REACTOR.getRecipes(method_10997())) {
			if (validateRecipe((FusionReactorRecipe) recipe)) {
				currentRecipe = (FusionReactorRecipe) recipe;
				crafingTickTime = 0;
				finalTickTime = currentRecipe.getTime();
				neededPower = currentRecipe.getStartEnergy();
				hasStartedCrafting = false;
				break;
			}
		}
	}
	
	/**
	 * Validates if reactor has all inputs and can output result
	 * 
	 * @param recipe
	 * @return
	 */
	private boolean validateRecipe(FusionReactorRecipe recipe) {
		return hasAllInputs(recipe) && canFitStack(recipe.getOutputs().get(0), outputStackSlot, true);	
	}
	
	/**
	 * Check if BlockEntity has all necessary inputs for recipe provided
	 * 
	 * @param recipeType RebornRecipe Recipe to check inputs
	 * @return Boolean True if reactor has all inputs for recipe
	 */
	private boolean hasAllInputs(RebornRecipe recipeType) {
		if (recipeType == null) {
			return false;
		}
		for (RebornIngredient ingredient : recipeType.getRebornIngredients()) {
			boolean hasItem = false;
			if (ingredient.test(inventory.method_5438(topStackSlot))
					|| ingredient.test(inventory.method_5438(bottomStackSlot))) {
				hasItem = true;
			}
			if (!hasItem) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Decrease stack size on the given slot according to recipe input
	 * 
	 * @param slot int Slot number
	 */
	private void useInput(int slot) {
		if (currentRecipe == null) {
			return;
		}
		for (RebornIngredient ingredient : currentRecipe.getRebornIngredients()) {
			if (ingredient.test(inventory.method_5438(slot))) {
				inventory.shrinkSlot(slot, ingredient.getCount());
				break;
			}
		}
	}


	// TilePowerAcceptor
	@Override
	public void method_16896() {
		super.method_16896();

		if (field_11863.field_9236) {
			return;
		}

		if(lastTick == field_11863.method_8510()){
			//Prevent tick accerators, blame obstinate for this.
			return;
		}
		lastTick = field_11863.method_8510();

		// Force check every second
		if (field_11863.method_8510() % 20 == 0) {
			checkCoils();
			inventory.setChanged();
		}

		if (coilCount == 0) {
			resetCrafter();
			return;
		}

		if (currentRecipe == null && inventory.hasChanged()) {
			updateCurrentRecipe();
		}

		if (currentRecipe != null) {
			if (!hasStartedCrafting && inventory.hasChanged() && !validateRecipe(currentRecipe)) {
				resetCrafter();
				return;
			}

			if (!hasStartedCrafting) {
				// Ignition!
				if (canUseEnergy(currentRecipe.getStartEnergy())) {
					useEnergy(currentRecipe.getStartEnergy());
					hasStartedCrafting = true;
					useInput(topStackSlot);
					useInput(bottomStackSlot);
				}
			}
			if (hasStartedCrafting && crafingTickTime < finalTickTime) {
				crafingTickTime++;
				// Power gen
				if (currentRecipe.getPower() > 0) {
					// Waste power if it has no where to go
					addEnergy(currentRecipe.getPower() * getPowerMultiplier());
					powerChange = currentRecipe.getPower() * getPowerMultiplier();
				} else { // Power user
					if (canUseEnergy(currentRecipe.getPower() * -1)) {
						setEnergy(getEnergy() - currentRecipe.getPower() * -1);
					}
				}
			} else if (crafingTickTime >= finalTickTime) {
				class_1799 result = currentRecipe.getOutputs().get(0);
				if (canFitStack(result, outputStackSlot, true)) {
					if (inventory.method_5438(outputStackSlot).method_7960()) {
						inventory.method_5447(outputStackSlot, result.method_7972());
					} else {
						inventory.shrinkSlot(outputStackSlot, -result.method_7947());
					}
					if (validateRecipe(this.currentRecipe)) {
						crafingTickTime = 0;
						useInput(topStackSlot);
						useInput(bottomStackSlot);
					} else {
						resetCrafter();
					}
				}
			}
			method_5431();
		}

		inventory.resetChanged();
	}


	@Override
	public double getPowerMultiplier() {
		double calc = (1F/2F) * Math.pow(size -5, 1.8);
		return Math.max(Math.round(calc * 100D) / 100D, 1D);
	}

	@Override
	public double getBaseMaxPower() {
		return Math.min(TechRebornConfig.fusionControlComputerMaxEnergy * getPowerMultiplier(), Integer.MAX_VALUE);
	}

	@Override
	public boolean canAcceptEnergy(class_2350 direction) {
		return !(direction == class_2350.field_11033 || direction == class_2350.field_11036);
	}

	@Override
	public boolean canProvideEnergy(class_2350 direction) {
		return direction == class_2350.field_11033 || direction == class_2350.field_11036;
	}

	@Override
	public double getBaseMaxOutput() {
		if (!hasStartedCrafting) {
			return 0;
		}
		return Integer.MAX_VALUE;
	}

	@Override
	public double getBaseMaxInput() {
		if (hasStartedCrafting) {
			return 0;
		}
		return TechRebornConfig.fusionControlComputerMaxInput;
	}

	@Override
	public void method_11014(final class_2487 tagCompound) {
		super.method_11014(tagCompound);
		this.crafingTickTime = tagCompound.method_10550("crafingTickTime");
		this.finalTickTime = tagCompound.method_10550("finalTickTime");
		this.neededPower = tagCompound.method_10550("neededPower");
		this.hasStartedCrafting = tagCompound.method_10577("hasStartedCrafting");
		if(tagCompound.method_10545("hasActiveRecipe") && tagCompound.method_10577("hasActiveRecipe") && this.currentRecipe == null){
			for (final RebornRecipe reactorRecipe : ModRecipes.FUSION_REACTOR.getRecipes(method_10997())) {
				if (validateRecipe((FusionReactorRecipe) reactorRecipe)) {
					this.currentRecipe = (FusionReactorRecipe) reactorRecipe;
				}
			}
		}
		if(tagCompound.method_10545("size")){
			this.size = tagCompound.method_10550("size");
		}
		this.size = Math.min(size, TechRebornConfig.fusionControlComputerMaxCoilSize);//Done here to force the samller size, will be useful if people lag out on a large one.
	}

	@Override
	public class_2487 method_11007(final class_2487 tagCompound) {
		super.method_11007(tagCompound);
		tagCompound.method_10569("crafingTickTime", this.crafingTickTime);
		tagCompound.method_10569("finalTickTime", this.finalTickTime);
		tagCompound.method_10569("neededPower", this.neededPower);
		tagCompound.method_10556("hasStartedCrafting", this.hasStartedCrafting);
		tagCompound.method_10556("hasActiveRecipe", this.currentRecipe != null);
		tagCompound.method_10569("size", size);
		return tagCompound;
	}

	// TileLegacyMachineBase
	@Override
	public void onLoad() {
		super.onLoad();
		this.checkCoils();
	}

	@Override
	public boolean canBeUpgraded() {
		return false;
	}

	// IToolDrop
	@Override
	public class_1799 getToolDrop(class_1657 playerIn) {
		return TRContent.Machine.FUSION_CONTROL_COMPUTER.getStack();
	}

	// ItemHandlerProvider
	@Override
	public RebornInventory<FusionControlComputerBlockEntity> getInventory() {
		return inventory;
	}

	// IContainerProvider
	@Override
	public BuiltContainer createContainer(int syncID, final class_1657 player) {
		return new ContainerBuilder("fusionreactor").player(player.field_7514).inventory().hotbar()
				.addInventory().blockEntity(this).slot(0, 34, 47).slot(1, 126, 47).outputSlot(2, 80, 47).syncEnergyValue()
				.syncIntegerValue(this::getCoilStatus, this::setCoilStatus)
				.syncIntegerValue(this::getCrafingTickTime, this::setCrafingTickTime)
				.syncIntegerValue(this::getFinalTickTime, this::setFinalTickTime)
				.syncIntegerValue(this::getSize, this::setSize)
				.syncIntegerValue(this::getState, this::setState)
				.syncIntegerValue(this::getNeededPower, this::setNeededPower).addInventory().create(this, syncID);
	}

	public int getCoilStatus() {
		return coilCount;
	}

	public void setCoilStatus(int coilStatus) {
		this.coilCount = coilStatus;
	}

	public int getCrafingTickTime() {
		return crafingTickTime;
	}

	public void setCrafingTickTime(int crafingTickTime) {
		this.crafingTickTime = crafingTickTime;
	}

	public int getFinalTickTime() {
		return finalTickTime;
	}

	public void setFinalTickTime(int finalTickTime) {
		this.finalTickTime = finalTickTime;
	}

	public int getNeededPower() {
		return neededPower;
	}

	public void setNeededPower(int neededPower) {
		this.neededPower = neededPower;
	}

	public int getSize() {
		return size;
	}

	public void setSize(int size) {
		this.size = size;
	}

	public void changeSize(int sizeDelta){
		int newSize = size + sizeDelta;
		this.size = Math.max(6, Math.min(TechRebornConfig.fusionControlComputerMaxCoilSize, newSize));
	}

	public int getState(){
		if(currentRecipe == null ){
			return 0; //No Recipe
		}
		if(!hasStartedCrafting){
			return 1; //Waiting on power
		}
		if(hasStartedCrafting){
			return 2; //Crafting
		}
		return -1;
	}

	public void setState(int state){
		this.state = state;
	}

	public String getStateString(){
		if(state == -1){
			return "";
		} else if (state == 0){
			return "No recipe";
		} else if (state == 1){
			return "Charging";
		} else if (state == 2){
			return "Crafting";
		}
		return "";
	}
}
